int exportAudioCursor;
int exportAudioDuration = 60 * 15;
int exportAudioSilence = 3;


void wr_dword_lh(unsigned char* data, unsigned int val)
{
	data[0] = val & 0xff;
	data[1] = (val >> 8) & 0xff;
	data[2] = (val >> 16) & 0xff;
	data[3] = (val >> 24) & 0xff;
}



void wr_word_lh(unsigned char* data, unsigned short val)
{
	data[0] = val & 0xff;
	data[1] = (val >> 8) & 0xff;
}



void set_export_percent(int percent)
{
	static int percentPrev = -1;

	if (percentPrev != percent)
	{
		percentPrev = percent;

		backgroundTaskDonePercent = percent;

		SDL_Event userEvent;

		userEvent.type = SDL_USEREVENT;
		userEvent.user.code = EVENT_SCREEN_UPDATE;
		userEvent.user.data1 = 0;
		userEvent.user.data2 = 0;

		SDL_PushEvent(&userEvent);
	}
}



//actual WAV export is done in a background thread, as it may take a while, with a GUI freeze

int exth_duration = 0;
const char* exth_filename = NULL;
int exth_detect_silence = 0;

int export_audio_to_file_thread(void* param)
{
	unsigned char header[44];

	int channels = 2;
	int block_duration = 1;//seconds
	int sample_len = exth_duration * SAMPLE_RATE;
	int sample_bytes = sample_len * channels * sizeof(short int);
	int file_size = sizeof(header) + sample_bytes;

	memcpy(header, "RIFF", 4);			//riff signature
	wr_dword_lh(&header[4], file_size - 8);//file size-8
	memcpy(&header[8], "WAVEfmt ", 8);	//format signature
	wr_dword_lh(&header[16], 16);		//header size
	wr_word_lh(&header[20], 1);			//uncompressed PCM
	wr_word_lh(&header[22], channels);			//stereo
	wr_dword_lh(&header[24], SAMPLE_RATE);
	wr_dword_lh(&header[28], SAMPLE_RATE * channels * sizeof(short int));//byterate (sample rate * channels * bytes per sample)
	wr_dword_lh(&header[32], 16 / 8);			//bytes per sample
	wr_dword_lh(&header[34], 16);		//bits per sample
	memcpy(&header[36], "data", 4);		//data signature
	wr_dword_lh(&header[40], sample_bytes);//sample data size in bytes

	PlayLayer* player_render = new PlayLayer();

	player_render->Play(scriptGeneratedData1, scriptGeneratedData1Size, SAMPLE_RATE);
	
	FILE* file = xfopen(exth_filename, "wb");

	if (file)
	{
		xfwrite(header, sizeof(header), 1, file);

		int samples_to_render = sample_len;
		int samples_rendered = 0;
		int block_silence_cnt = 0;

		int block_samples = SAMPLE_RATE * block_duration;
		int block_bytes = block_samples * channels * sizeof(short int);

		short int* buf = (short int*)malloc(block_bytes);

		while (samples_to_render > 0)
		{
			set_export_percent((samples_rendered / SAMPLE_RATE) * 100 / (sample_len / SAMPLE_RATE));

			if (samples_to_render < block_samples) block_samples = samples_to_render;

			memset(buf, 0, block_bytes);

			player_render->Render(buf, block_samples * channels);

			if (exth_detect_silence > 0)
			{
				long acc = 0;

				for (int i = 0; i < block_samples * channels; ++i) acc += labs(buf[i]);

				if (acc < (SAMPLE_RATE * channels * block_duration))
				{
					++block_silence_cnt;

					if (block_silence_cnt >= exth_detect_silence)
					{
						//printf("stop at %i (acc=%ul)\n", samples_rendered / SAMPLE_RATE, acc);
						break;
					}
				}
			}

			xfwrite(buf, block_bytes, 1, file);

			samples_to_render -= block_samples;
			samples_rendered += block_samples;
		}

		free(buf);

		set_export_percent(100);

		xfclose(file);

		if (exth_detect_silence > 0)
		{
			//update file header to actual size

			file = xfopen(exth_filename, "r+b");

			if (file)
			{
				xfseek(file, 0, SEEK_SET);
				xfread(header, sizeof(header), 1, file);

				sample_bytes = samples_rendered * channels * sizeof(short int);
				file_size = sizeof(header) + sample_bytes;

				wr_dword_lh(&header[4], file_size - 8);//file size-8
				wr_dword_lh(&header[40], sample_bytes);//sample data size in bytes

				xfseek(file, 0, SEEK_SET);
				xfwrite(header, sizeof(header), 1, file);
				xfclose(file);
			}
		}
	}

	delete player_render;

	script_free_data1();

	backgroundTaskActive = false;
	set_export_percent(-1);

	return 0;
}



//duration in seconds, detect silence in seconds to cut early before the duration expires (zero disables detection)

bool export_audio_to_file(const char* filename, int duration, int detect_silence)
{
	//compile and play

	music_stop();

	if (!compile_song(0, 0, false, true, false)) return false;

	if (!scriptGeneratedData1 || !scriptGeneratedData1Size) return false;

	//start export thread

	exth_duration = duration;
	exth_filename = filename;
	exth_detect_silence = detect_silence;

	backgroundTaskActive = true;
	backgroundTaskDonePercent = 0;

	SDL_Thread *thread = SDL_CreateThread(export_audio_to_file_thread, "wav_export", NULL);

	return true;
}



void set_export_audio(void)
{
	music_stop();

	set_edit_mode(MODE_EXPORT_AUDIO);

	exportAudioCursor = 0;
}


void draw_export_audio(void)
{
	char buf[32];

	for (int i = 0; i < textScreenWidth; ++i)
	{
		PUT_ATTR(i,0,COL_HEADER);
		PUT_ATTR(i,textScreenHeight-1,COL_HEADER);
	}

	put_str(1,0,"Export song audio",COL_HEADER);

	int y = PATTERN_YPOS + 1;

	put_str(PATTERN_XPOS + 1, y, " Max export duration ", (exportAudioCursor == 0) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND);

	snprintf(buf, sizeof(buf), "%4.4u sec", exportAudioDuration);

	put_str(PATTERN_XPOS + 30, y, buf, COL_PROGRAM_BACKGROUND);

	y += 2;

	put_str(PATTERN_XPOS + 1, y, " Cut at a period of silence ", (exportAudioCursor == 1) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND);

	snprintf(buf, sizeof(buf), "%2.2u sec (0 to disable)", exportAudioSilence);

	put_str(PATTERN_XPOS + 30, y, buf, COL_PROGRAM_BACKGROUND);

	y += 2;

	put_str(PATTERN_XPOS + 1, y, " Export audio file (*.wav) ", (exportAudioCursor == 2) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND);
}



void export_audio_events(SDL_Event event)
{
	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			switch(event.key.keysym.sym)
			{
			case SDLK_ESCAPE:
				{
					set_edit_mode(MODE_EDIT_SONG);
				}
				break;

			case SDLK_UP:
				{
					--exportAudioCursor;

					if (exportAudioCursor < 0) exportAudioCursor = 2;
				}
				break;

			case SDLK_DOWN:
				{
					++exportAudioCursor;

					if (exportAudioCursor > 2) exportAudioCursor = 0;
				}
				break;

			case SDLK_PAGEUP:
			{
				exportAudioCursor = 0;
			}
			break;

			case SDLK_PAGEDOWN:
			{
				exportAudioCursor = 2;
			}
			break;

			case SDLK_RETURN:
				{
					switch (exportAudioCursor)
					{
					case 0: set_input_number(PATTERN_XPOS + 30, PATTERN_YPOS + 1, 4, MODE_EXPORT_AUDIO, true, &exportAudioDuration, false, NULL); break;
					case 1: set_input_number(PATTERN_XPOS + 30, PATTERN_YPOS + 3, 2, MODE_EXPORT_AUDIO, true, &exportAudioSilence, false, NULL); break;
					case 2:
						{
							set_select_file("Export audio file", exportDirectory, sizeof(exportDirectory), MODE_EDIT_SONG, ".wav", true, EVENT_EXPORT_AUDIO_FILE, selectFileImportLast, sizeof(selectFileImportLast), 1, 0);
						}
						break;
					}
				}
				break;
			}
		}
		break;
	}
}