int inputX;
int inputY;
int inputMaxLen;
int inputType;
int inputNewNumber;
int inputReturnMode;
bool inputClearOnFirstChange;
int inputCursor;
char inputNewString[1024];
int* inputOriginalNumber;
char* inputOriginalString;
bool inputNumberSigned;
bool inputDone;
void(*inputDoneCB)(void) = NULL;

#define INPUT_TYPE_STRING		0
#define INPUT_TYPE_NUMBER		1
#define INPUT_TYPE_HEX_STRING	2



void set_input_number(int x, int y, int len, int mode, bool changec, int* number, bool sign, void(*doneCB)(void))
{
	memset(inputNewString, 0, sizeof(inputNewString));

	inputNewNumber = *number;
	inputOriginalNumber = number;

	inputType = INPUT_TYPE_NUMBER;
	inputX = x;
	inputY = y;
	inputMaxLen = len;
	inputReturnMode = mode;
	inputClearOnFirstChange = changec;
	inputNumberSigned = sign;
	inputDoneCB = doneCB;

	set_edit_mode(MODE_INPUT);
}



int input_actual_string_length(void)
{
	for (int i = inputMaxLen - 1; i >= 0; --i)
	{
		if (inputNewString[i] > ' ') return i + 1;
	}

	return 0;
}


int input_max_string_length(void)
{
	return inputMaxLen;
}



void set_input_string(int x, int y, int max_len, int mode, bool changec, char* string, void(*doneCB)(void))
{
	memset(inputNewString, 0, sizeof(inputNewString));
	strncpy(inputNewString, string, sizeof(inputNewString));

	inputOriginalString = string;

	inputType = INPUT_TYPE_STRING;
	inputX = x;
	inputY = y;
	inputMaxLen = max_len;
	inputReturnMode = mode;
	inputClearOnFirstChange = changec;
	inputCursor = input_actual_string_length();
	inputDoneCB = doneCB;

	set_edit_mode(MODE_INPUT);
}



void set_input_hex_string(int x,int y,int max_len,int mode,char* string)
{
	memset(inputNewString,0,sizeof(inputNewString));
	strncpy(inputNewString,string,sizeof(inputNewString));

	inputOriginalString=string;

	inputType=INPUT_TYPE_HEX_STRING;
	inputX=x;
	inputY=y;
	inputMaxLen=max_len;
	inputReturnMode=mode;
	inputCursor=0;

	set_edit_mode(MODE_INPUT);
}



void draw_input(void)
{
	switch(inputType)
	{
	case INPUT_TYPE_NUMBER:
		{
		if (inputNumberSigned) PUT_CHAR(inputX - 1, inputY, (inputNewNumber < 0) ? '-' : ' ');

		put_num_attr(inputX, inputY, abs(inputNewNumber), inputMaxLen, COL_CURSOR);
		}
		break;

	case INPUT_TYPE_STRING:
		{
			int len = inputMaxLen;
			int off = 0;

			if (len > (textScreenWidth - inputX - 2))
			{
				len = textScreenWidth - inputX - 2;
				off = inputCursor - len + 1;
				if (off < 0) off = 0;

				if (off > 0)
				{
					PUT_CHAR_ATTR(inputX - 1, inputY, '<', COL_PROGRAM_BACKGROUND);
				}
				else
				{
					PUT_CHAR_ATTR(inputX - 1, inputY, ' ', COL_PROGRAM_BACKGROUND);
				}

				if ((int)strlen(&inputNewString[off]) > len)
				{
					PUT_CHAR_ATTR(inputX + inputMaxLen - 2, inputY, '>', COL_PROGRAM_BACKGROUND);
				}
				else
				{
					PUT_CHAR_ATTR(inputX + inputMaxLen - 2, inputY, ' ', COL_PROGRAM_BACKGROUND);
				}
			}

			for (int i = 0; i < len; ++i)
			{
				char c = inputNewString[off];
				if (c < 0x20) c = 0x20;
				PUT_CHAR_ATTR(inputX + i, inputY, c, (off == inputCursor) ? COL_CURSOR : COL_CURSOR_ROW);
				++off;
			}
		}
		break;

	case INPUT_TYPE_HEX_STRING:
		{
			int len = inputMaxLen;
			int off = 0;

			if (len > (textScreenWidth - inputX - 2))
			{
				len = textScreenWidth - inputX - 2;
				off = inputCursor - len + 1;
				if (off < 0) off = 0;

				if (off > 0)
				{
					PUT_CHAR_ATTR(inputX - 1, inputY, '<', COL_PROGRAM_BACKGROUND);
				}
				else
				{
					PUT_CHAR_ATTR(inputX - 1, inputY, ' ', COL_PROGRAM_BACKGROUND);
				}

				if ((int)strlen(&inputNewString[off]) > len)
				{
					PUT_CHAR_ATTR(textScreenWidth - 2, inputY, '>', COL_PROGRAM_BACKGROUND);
				}
				else
				{
					PUT_CHAR_ATTR(textScreenWidth - 2, inputY, ' ', COL_PROGRAM_BACKGROUND);
				}
			}

			for (int i = 0; i < len; ++i)
			{
				PUT_CHAR_ATTR(inputX + i, inputY, NUM2HEX(inputNewString[off]), (off == inputCursor) ? COL_CURSOR : COL_CURSOR_ROW);
				++off;
			}
		}
		break;
	}
}



void input_events(SDL_Event event)
{
	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			int key = event.key.keysym.sym;

			switch(inputType)
			{
			case INPUT_TYPE_NUMBER:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							*inputOriginalNumber=inputNewNumber;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							inputNewNumber/=10;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_DELETE:
						{
							inputNewNumber=0;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_MINUS:
						{
							if(inputNumberSigned)
							{
								inputNewNumber=0-inputNewNumber;
							}
						}
						break;
					}

					if(key>=SDLK_0&&key<=SDLK_9)
					{
						if(inputClearOnFirstChange)
						{
							inputNewNumber=0;
							inputClearOnFirstChange=false;
						}

						bool sign = inputNewNumber < 0 ? true : false;

						inputNewNumber=(abs(inputNewNumber)*10+(key-SDLK_0))%(int)pow(10.0f,inputMaxLen);

						if(sign) inputNewNumber=0-inputNewNumber;
					}
				}
				break;

			case INPUT_TYPE_STRING:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							inputNewString[input_actual_string_length()] = 0;
							strncpy(inputOriginalString, inputNewString, input_max_string_length());
							
							if (inputDoneCB != NULL) inputDoneCB();

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							if(inputCursor)
							{
								memcpy(&inputNewString[inputCursor - 1], &inputNewString[inputCursor], inputMaxLen - inputCursor);
								inputNewString[inputMaxLen - 1] = 0;

								--inputCursor;
							}

							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_LEFT:
						{
							if(inputCursor) --inputCursor;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_RIGHT:
						{
							if(inputCursor<input_actual_string_length()) ++inputCursor;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_HOME:
						{
							inputCursor=0;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_END:
						{
							inputCursor=input_actual_string_length();
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_INSERT:
					{
						memcpy(&inputNewString[inputCursor + 1], &inputNewString[inputCursor], inputMaxLen - (inputCursor + 1));
						inputNewString[inputCursor] = ' ';
						inputClearOnFirstChange = false;
					}
					break;
					
					case SDLK_DELETE:
						{
						memcpy(&inputNewString[inputCursor], &inputNewString[inputCursor + 1], inputMaxLen - (inputCursor + 1));
							inputNewString[inputMaxLen - 1] = 0;
							inputClearOnFirstChange=false;
							key = 0;
						}
						break;
					}

					if (key >= 32 && key < 128)
					{
						if (SDL_GetModState() & KMOD_SHIFT)
						{
							if (key >= 'a'&&key <= 'z')  key -= 0x20;
							if (key == '1') key = '!';
							if (key == '2') key = '@';
							if (key == '3') key = '#';
							if (key == '4') key = '$';
							if (key == '5') key = '%';
							if (key == '6') key = '^';
							if (key == '7') key = '&';
							if (key == '8') key = '*';
							if (key == '9') key = '(';
							if (key == '0') key = ')';
							if (key == '-') key = '_';
							if (key == '=') key = '+';
							if (key == '\\') key = '|';
							if (key == '[') key = '{';
							if (key == ']') key = '}';
							if (key == ';') key = ':';
							if (key == '\'') key = '\"';
							if (key == ',') key = '<';
							if (key == '.') key = '>';
							if (key == '/') key = '?';
							if (key == '`') key = '~';
						}

						if(inputClearOnFirstChange)
						{
							memset(inputNewString,0,sizeof(inputNewString));
							inputClearOnFirstChange=false;
							inputCursor=0;
						}

						if(input_actual_string_length()<inputMaxLen-1)
						{
							memcpy(&inputNewString[inputCursor+1],&inputNewString[inputCursor],sizeof(inputNewString)-inputCursor-1);

							inputNewString[inputCursor++]=key;

							if(inputCursor>inputMaxLen-1) inputCursor=inputMaxLen-1;
						}
					}
				}
				break;

			case INPUT_TYPE_HEX_STRING:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							memcpy(inputOriginalString,inputNewString,inputMaxLen);

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							inputNewString[inputCursor]=0;

							if(inputCursor) --inputCursor;
						}
						break;

					case SDLK_LEFT:
						{
							if(inputCursor) --inputCursor;
						}
						break;

					case SDLK_RIGHT:
						{
							if(inputCursor<inputMaxLen-1) ++inputCursor;
						}
						break;

					case SDLK_HOME:
						{
							inputCursor=0;
						}
						break;

					case SDLK_END:
						{
							inputCursor=inputMaxLen-1;
						}
						break;

					case SDLK_DELETE:
						{
							inputNewString[inputCursor]=0;

							if(inputCursor<inputMaxLen-1) ++inputCursor;
						}
						break;
					}

					if(key>=SDLK_0&&key<=SDLK_9)
					{
						inputNewString[inputCursor]=key-SDLK_0;

						if(inputCursor<inputMaxLen-1) ++inputCursor;
					}

					if(key>=SDLK_a&&key<=SDLK_f)
					{
						inputNewString[inputCursor]=key-SDLK_a+10;

						if(inputCursor<inputMaxLen-1) ++inputCursor;
					}
				}
				break;
			}
		}
		break;
	}
}